/*****************************************************************************\
*                                                                             *
* 				  SES.h protocol definitions                  *
*                                                                             *
\*****************************************************************************/
/* included both in Klone (klsesi.c) and in extensions (seslib.c)
 * This defines the protocol data themselves, independently on the C or Klone
 * mappings
 */

#ifndef SES_H
#define SES_H

#define NO_SHM 1			/* temporary till it works */

#ifndef NO_SHM				/* SHM is more readable in .c files */
#define SHM 1
#endif
					/* version of the protocol */
#define KlSesVersion 0

					/* transport layer types */
#define KlSesTransportPipe 0
#define KlSesTransportSHM 1

/* find the granularity of padding, the double is the most sensitive type */
typedef struct {char c; double d;}  KlSESDummy;
#define SES_padlen (sizeof(KlSESDummy) - sizeof(double)) /* grain */
#define SES_padded(l) (((((l)-1)/SES_padlen)+1)*SES_padlen)	/* padded to */
#define SES_padded_structlen(name) SES_padded(sizeof(struct name))

/*****************************************************************************\
* 			 protocol structures: SES_PS                          *
\*****************************************************************************/

typedef unsigned char uchar;

typedef struct _SES_PS_Header {		/* the header of all requests */
    uchar type;				/* SES_HT_... */
    uchar module_id;			/* the extension module id */
    unsigned int length;		/* total length in bytes incl. head. */
    struct _KlSesReqId *reqid;	        /* sequence id of this request */
    union {				/* what function to apply, can be: */
	void *ptr;			/*   function opaque ID, or */
	unsigned int id;		/*   ID, or seq. number replied to */
    } func;
    int arity;				/* number of arguments after func */
    unsigned int header_checksum;       /* simple sum, see def below */
} *SES_PS_Header;

#define SES_PS_Header_padlen SES_padded_structlen(_SES_PS_Header)
#define SES_PS_HeaderEnd(h)  (((uchar *) (h)) + SES_PS_Header_padlen)
#define SES_PS_HeaderBodyEnd(h)  \
    (((uchar *) (h)) + SES_PS_Header_padlen + (h)->length)
#define SES__HeaderChecksum(request) 					\
    ((unsigned int) request->type + (unsigned int) request->module_id + \
     (unsigned int) request->length + (unsigned int) request->reqid + \
     (unsigned int) request->func.ptr + (unsigned int) request->arity)

typedef struct _SES_PS_Any {		/* common part */
    uchar type;
} *SES_PS_Any;

typedef struct _SES_PS_Int {		/* integer */
    uchar type;
    int number;
} *SES_PS_Int;
#define SES_PS_Int_padlen SES_padded_structlen(_SES_PS_Int)

typedef struct _SES_PS_Boolean {	/* boolean */
    uchar type;
    int number;
} *SES_PS_Boolean;
#define SES_PS_Boolean_padlen SES_padded_structlen(_SES_PS_Boolean)

typedef struct _SES_PS_Ptr {		/* void * */
    uchar type;
    void *ptr;
} *SES_PS_Ptr;
#define SES_PS_Ptr_padlen SES_padded_structlen(_SES_PS_Ptr)

typedef struct _SES_PS_String {		/* string */
    uchar type;
    int length;
    uchar string[1];			/* the string itself */
} *SES_PS_String;
/* s is length of string (room for 0 is added) */
#define SES_PS_String_padlen(s) SES_padded(sizeof(struct _SES_PS_String) + (s))

typedef struct _SES_PS_Atom {		/* atom */
    uchar type;
    int length;
    uchar string[1];			/* the string itself */
} *SES_PS_Atom;
/* s is length of string (room for 0 is added) */
#define SES_PS_Atom_padlen(s) SES_padded(sizeof(struct _SES_PS_Atom) + (s))

typedef struct _SES_PS_Real {		/* doubles */
    uchar type;
    double number;
} *SES_PS_Real;
#define SES_PS_Real_padlen SES_padded_structlen(_SES_PS_Real)

typedef struct _SES_PS_Id {		/* extension-defined opaque type */
    uchar type;
    int idtype;
    void *value;
} *SES_PS_Id;
#define SES_PS_Id_padlen SES_padded_structlen(_SES_PS_Id)


/*****************************************************************************\
* 			       protocol values                                *
\*****************************************************************************/
/* warning: this is INTERNAL to the protocol, private use only 
 * a general trick is to never have 0 as a valid value, to trap errors 
 */

/* SES_HT : HeaderTypes of requests. (SES_PS_Header->type). major numbers 
 * must be < 64
 */
					/* _Out_ main->ext */
#define SES_HT_Out_Call 1
#define SES_HT_Out_Send 2
#define SES_HT_Out_Result 3
#define SES_HT_Out_Error 4
#define SES_HT_Out_Sync 5
#define SES_HT_Out_Free 6
#define SES_HT_Out_Err
					/* _In_ ext->main */
#define SES_HT_In_Call 9
#define SES_HT_In_Send 10
#define SES_HT_In_Result 11
#define SES_HT_In_Error 12
#define SES_HT_In_Sync 13
#define SES_HT_In_Free 14	/* unused */

/* SES_HFI : Header Function Ids, predefined func_id numbers. minor numbers */

#define SES_HFI_ReadEvalText 1		/* read & eval the following text */
#define SES_HFI_Ready 2			/* end of Inits, you can use it now */
#define SES_HFI_DeclareFunction 3	/* declare a function */
#define SES_HFI_DeclareType 4		/* declare a type */
#define SES_HFI_NotFound 5		/* fatal error, extension not found */

#define SES_HFI_ErrorResult 0
#define SES_HFI_ErrorHandlerResult 1

/* API data types. you need to update also codecs */
#define SES_None 0
#define SES_Int 1
#define SES_String 2
#define SES_Atom 3
#define SES_Id 4
#define SES_Boolean 5
#define SES_Real 6
#define SES_Ptr 7
#define SES_ERROR 8
#define SES_MAX_Builtins 9		/* first unused API data type */
    /* Id types will be appended dynamically here... */

#endif /* SES_H */
