/*****************************************************************************\
*                                                                             *
* 				  SES lib                                     *
*                                                                             *
\*****************************************************************************/
/* Generic support to be linked in each C extension
 * SES_XXX symbols are externally usable, SES__XXX should stay private to 
 * this lib
 * could also used in the main Klone interpreter, where we suppose that KLSES
 * is #defined
 * the shared library code can be undefined by defining NO_SHM
 */

#include <stdio.h>
#include <fcntl.h>
#include <errno.h>
#ifdef NEED_SELECT_H
#include <sys/select.h>
#endif
#include <sys/time.h>

#include "seslib.h"			/* includes also ses.h */
#include "kclist.c"			/* includes also kclist.h */

#include "sescodecs.c"			/* codecs, common to both sides */

					/* imports */
extern char *getenv _P_((char *));

					/* variables */
#ifndef SES_BUFSIZE
#define SES_BUFSIZE 16 * 1024		/* 16k buffer default value for socket */
#endif

#define ECHECK(err) (errno == err)
#define ESET(val) errno = val

#if defined(SYSV) || defined(SVR4)
#define KlNDELAY O_NDELAY
#else
#define KlNDELAY FNDELAY
#endif

/* internal variables */

static int SES__id_last = SES_MAX_Builtins;
static SES_Connection SES__CurrentSescon; /* for error messages */

/* internal functions forward declarations, prefixed with SES__ */

static int SES__ReadWhatsAvailable _P_((SES_Connection sescon));
static void SES__DecodeRequest _P_((SES_Connection sescon, 
				    SES_PS_Header request));
static void SES__DecodeRequestArgs _P_((SES_Connection sescon,
					SES_PS_Header request, SES_Args args));
static SES_PS_Header SES__NewRequest _P_((SES_Connection sescon,
					  unsigned int type, 
					  unsigned int func_id, void *reqid,
					  unsigned int size, int arity));
static SES_PS_Header SES__AppendToRequest _P_((SES_Connection sescon, 
					SES_PS_Header request,
					unsigned int size, uchar *buffer));
static void SES__WaitForErrorHandling _P_((SES_Args args));

#define SES__RequestPending(sescon) \
    (SES_PS_HeaderEnd(sescon->inptr) <= sescon->inend \
     && (sescon->inptr + ((SES_PS_Header) sescon->inptr)->length) \
         <= sescon->inend)

/* validates (recomputes) the checksum of a Request
 * needed after a SES__AppendToRequest, but not after a SES__NewRequest
 */
#define SES__ValidateRequest(request) \
    request->header_checksum = SES__HeaderChecksum(request)


/*****************************************************************************\
* 			   fatal connection errors                            *
\*****************************************************************************/

/* the (modifiable) table of error messages. note that these are all fatal 
 * errors, so the extension will terminate immediately afterwards 
 */

int SES_errno = 0;
char *SES_errtexts[] = {
    "UNKNOWN ERROR",			/* 0: should not happen */
    "extension not started properly", /* 1 */
    "connection closed with main program", /* 2 */
    "connection broken",		/* 3 */
    "internal: protocol error", /* 4 */
    "internal: invalid function ID", /* 5 */
    "internal: bad argument type", /* 6 */
    "extension: return value omitted when answering a call", /* 7 */
    "bad function declaration", /* 8 */
    "extension building ovelapping requests", /* 9 */
    0
};

void
SES_FatalIOErrorHandlerDefault(sescon, num)
    SES_Connection sescon;
    int num;
{
    fprintf(stderr, "SES extension: Fatal IO error #%d, aborting:\n    %s\n",
	    num, SES_errtexts[num]);
    exit(1);
}

void (*SES_FatalIOErrorHandler)_P_((SES_Connection sescon, int num))
    = SES_FatalIOErrorHandlerDefault;

SES__CODEC_ERROR(int expected_type, int provided_type)
{
    SES_FatalIOErrorHandler(SES__CurrentSescon, 4);
}

/*****************************************************************************\
* 				  Connection                                  *
\*****************************************************************************/

/* initialize structures of a new connection
 * the socket buffer is:
 */

void
SES_ConnectionInit(sescon)
    SES_Connection sescon;
{
    int buf_size = SES_BUFSIZE - sizeof(void *); /* room for malloc info */

    /* array of defined types of this module */
    sescon->IDs = KCListNew();
    /* allocate input and output buffers */
    sescon->inbuf = (uchar *) malloc(buf_size);
    sescon->inptr = sescon->inend = sescon->inbuf;
    sescon->inlimit = sescon->inbuf + buf_size;
    sescon->outbuf = (uchar *) malloc(buf_size);
    sescon->outptr = sescon->outbuf;
    sescon->outlimit = sescon->outbuf + buf_size;
}

/*****************************************************************************\
* 				   Handler                                    *
\*****************************************************************************/

/* Handler gets all it can into its buffer, and for each request stored, 
 * decode and execute it.
 * it only returns when there is no more complete request in the buffer
 */
void
SES_Handler(SES_Connection sescon)
{
    if (SES__ReadWhatsAvailable(sescon)) {
	while (SES__RequestPending(sescon)) {
	    SES__DecodeRequest(sescon, (SES_PS_Header) sescon->inptr);
	}
    } else {				/* select activated, but empty => EOF */
#ifdef SESDEBUG
	    fprintf(stderr, "SESLIB: connection cut, exiting normally\n");
#endif
	ESET(EPIPE);
	SES_FatalIOErrorHandler(sescon, 2);
    }
}


/* blocks till something comes, then parses it
 * can be used in a while loop while waiting for some condition, as it 
 * parses only one request at a time.
 */

void
SES_ReadAndHandle(SES_Connection sescon)
{
    if (SES__RequestPending(sescon)) {
	SES__DecodeRequest(sescon, (SES_PS_Header) sescon->inptr);
    } else {
	fd_set fds;
	FD_ZERO(&fds);
	FD_SET(sescon->fd_in, &fds);
	SES_MainPreSelect(sescon);
	select(sescon->fd_in+1, &fds, NULL, NULL, NULL);
	if (FD_ISSET(sescon->fd_in, &fds)) {
	    if (SES__ReadWhatsAvailable(sescon)) {
		if (SES__RequestPending(sescon)) {
		    SES__DecodeRequest(sescon, (SES_PS_Header) sescon->inptr);
		}
	    } else {		/* select activated, but empty => EOF */
#ifdef SESDEBUG
		fprintf(stderr, "SESLIB: connection cut, exiting normally\n");
#endif
		ESET(EPIPE);
		SES_FatalIOErrorHandler(sescon, 2);
	    }
	}
    }
}


/*****************************************************************************\
* 			     input queue handling                             *
\*****************************************************************************/

/* read all what is pending on the connection
 */

int					/* true if something was read */
SES__ReadWhatsAvailable(SES_Connection sescon)
{
    int size;

retry:
    if (size = (sescon->inend - sescon->inptr)) { /* move remnants at start */
	memmove(sescon->inbuf, sescon->inptr, size); /* may overlap */
    }
    sescon->inptr = sescon->inbuf;
    sescon->inend = sescon->inptr + size;
    if (sescon->inlimit - sescon->inend < sizeof (struct _SES_PS_Header)) {
	/* no room for a least another request, realloc */
	uchar *oldbuf = sescon->inbuf;
	size = 2 * (sescon->inlimit - sescon->inbuf) + sizeof(void *);
	sescon->inbuf = (uchar *) realloc(oldbuf, size);
	sescon->inptr = sescon->inbuf + (sescon->inptr - oldbuf);
	sescon->inlimit = sescon->inbuf + size;
	sescon->inend = sescon->inbuf + (sescon->inend - oldbuf);
    }
    /* append to the end of remnants already in buffer */
    size = read(sescon->fd_in, sescon->inend, sescon->inlimit - sescon->inend);
    if (size < 0) {			/* connection problem */
	if (!ECHECK(EINTR))		/* weird problem, quit */
	    SES_FatalIOErrorHandler(sescon, 3);
	else				/* normal interrupt, retry */
	    goto retry;
    }
    sescon->inend += size;
    return size;
}

/* processes all complete incoming requests
 * first, discriminates on request types
 */

void
SES__DecodeRequest(SES_Connection sescon, SES_PS_Header request)
{
    SES_PS_Header request_result;
    struct _SES_Args args_contents;
    SES_Args args = &args_contents;

    /* validity check. not really necessary on this side of the connection */
    if (request->header_checksum != SES__HeaderChecksum(request)) {
	SES_FatalIOErrorHandler(sescon, 4);
    }
    /* todo: better check of validity? */
    SES__DecodeRequestArgs(sescon, request, args);/* setup args/retval */
    sescon->inptr += request->length; /* goto next request */

    switch (request->type) {
    case SES_HT_Out_Call:
	request_result = SES__NewRequest(sescon, SES_HT_In_Result, 
					 0,
					 request->reqid,
					 SES_PS_Int_padlen, /* guess */
					 1);
	args->return_result = request_result;
	((SES_PS_Any) (((uchar *) request_result) + SES_PS_Header_padlen))
	    ->type = SES_None;	/* put a marker */

	(*((sescon_functions_type) request->func.ptr))(args); /* call */

	/* verify that a result has been returned, else error */
	if (((SES_PS_Any) SES_PS_HeaderEnd(args->return_result))->type 
	    == SES_None)
	    SES_FatalIOErrorHandler(sescon, 7);
	if (args->return_result->type == SES_HT_In_Error)
	    SES__WaitForErrorHandling(args);

	break;
    case SES_HT_Out_Send:		/* just execute function */
	(*((sescon_functions_type) request->func.ptr))(args); /* call */
	if (args->return_result	/* check error */
	    && args->return_result->type == SES_HT_In_Error)
	    SES__WaitForErrorHandling(args);
	break;
    case SES_HT_Out_Result:
	/* TODO */
	break;
    case SES_HT_Out_Free:	/* just apply free callback to data */
	(*((sescon_functions_type) request->func.ptr))(SES_ArgPtr(args));
	break;
    case SES_HT_Out_Error:
	if ((int) request->func.ptr) { /* error handler reply */
	    sescon->errhandler_reply = (int) request->func.ptr;
	} else {
	    /* TODO: called function returned error */
	}
	break;
    case SES_HT_Out_Sync:		/* synchronise on a reqid */
	/* send a void result for this seq. number, then flush */
	request_result = SES__NewRequest(sescon, SES_HT_In_Result, 0,
					 request->reqid, 0, 0);
	break;
        /* TODO: do asynch_call */
    default:
	SES_FatalIOErrorHandler(sescon, 4);
    }
}


/* initialize a pre-allocated args struct to first argument of a new request */

void
SES__DecodeRequestArgs(sescon, request, args)
    SES_Connection sescon;
    SES_PS_Header request;
    SES_Args args;
{
    args->ptr = ((uchar *) request) + SES_PS_Header_padlen;
    args->position = 0;
    args->sescon = sescon;
    args->request = request;
    args->return_result = 0;		/* by default, dont store retval */
}

/*****************************************************************************\
* 			    output queue handling                             *
\*****************************************************************************/

/* allocate a new output request
 * initialize its header, advances the queue structures, but just
 * reserves the room for parameters, to be filled by the caller
 */

SES_PS_Header				/* return the built request (headerp) */
SES__NewRequest(sescon, type, func_id, reqid, size, arity)
    SES_Connection sescon;
    unsigned int type;			/* SES_HT_... major # of request */
    unsigned int func_id;		/* id (minor) of request */
    void *reqid;			/* id of the (answered) request */
    unsigned int size;			/* size in bytes to reserve */
    int arity;				/* arity of request */
{
    SES_PS_Header request;
    /* pad size to granularity (debug code, normally size is already padded) */
    size = SES_padded(size);

					/* reserve size more bytes */
    if (sescon->outlimit - sescon->outptr </* remaining space too small*/
	(SES_PS_Header_padlen + size)) {	/* we must realloc */
	int newsize = sescon->outlimit - sescon->outbuf;
	uchar *oldbuf = sescon->outbuf;
	while ((newsize - (sescon->outlimit - sescon->outptr))
	       < (SES_PS_Header_padlen + size)) {
	    newsize = 2 * newsize + sizeof(void *);
	}
	sescon->outbuf = (uchar *) realloc(oldbuf, newsize);
	sescon->outptr = sescon->outbuf + (sescon->outptr - oldbuf);
	sescon->outlimit = sescon->outbuf + newsize;
    }
					/* encode header */
    request = (SES_PS_Header) sescon->outptr;
    request->type = type;
    request->module_id = sescon->id;
    request->func.id = func_id;
    request->length = SES_PS_Header_padlen + size;
    request->arity = arity;
    request->reqid = reqid;
					/* compute checksum */
    request->header_checksum = SES__HeaderChecksum(request);
					/* advance pointer */
    sescon->outptr += SES_PS_Header_padlen + size;

    return request;
}

/* appends data (size bytes) to a request. it calls SES__ValidateRequest,
 * so increment arity before alling it if you added a new argument
 * this will modify in place the length (+size)  of the request
 * returns the request, which may have changed due to a realloc
 * bytes in buffer will be appended to the end, except if buffer is a null ptr
 */

SES_PS_Header
SES__AppendToRequest(sescon, request, size, buffer)
    SES_Connection sescon;
    SES_PS_Header request;		/* the request */
    unsigned int size;			/* number of bytes to add */
    uchar *buffer;			/* the chars to add. 0 means no copy */
{
    size = SES_padded(size);		/* pad size to granularity (debug) */

					/* reserve size more bytes */
    if ((sescon->outlimit - sescon->outptr) /* remaining space too small*/
	< size) {	/* we must realloc */
	int newsize = sescon->outlimit - sescon->outbuf;
	uchar *oldbuf = sescon->outbuf;
	int offset = ((uchar *) request) - sescon->outbuf;
	while ((newsize - (sescon->outlimit - sescon->outptr))
	       < size) {
	    newsize = 2 * newsize + sizeof(void *);
	}
	sescon->outbuf = (uchar *) realloc(oldbuf, newsize);
	sescon->outptr = sescon->outbuf + (sescon->outptr - oldbuf);
	sescon->outlimit = sescon->outbuf + newsize;
	request = (SES_PS_Header) (sescon->outbuf + offset);
    }

    if ((((uchar *) request) + request->length) < sescon->outptr) { /* debug */
	/* cannot append, another request already there! */
	SES_FatalIOErrorHandler(SES__CurrentSescon, 9);
    }
    if (buffer)
	memcpy(((uchar *) request) + request->length, buffer, size);
    request->length += size;
    SES__ValidateRequest(request);
    sescon->outptr += size;
    return request;
}

void
SES_Flush(sescon)
    SES_Connection sescon;
{
    uchar *buffer = sescon->outbuf;
    int bytes_to_send = sescon->outptr - sescon->outbuf;
    while (bytes_to_send) {
	int sent;
	if (-1 == (sent = 
		   write(sescon->fd_out, buffer, bytes_to_send))) { /* error */
	    SES_FatalIOErrorHandler(sescon, 2);
	}
	bytes_to_send -= sent;
	buffer += sent;
    }
    sescon->outptr = sescon->outbuf;
}

/*****************************************************************************\
* 			       Declare Function                               *
\*****************************************************************************/
/* Declaring a function is sending the built-in function SES_HFI_DeclareFunction
 * with parameters:
 *   - func ptr
 *   - name
 *   - arity
 *   - return type
 *   - parameters types
 * current implementation limits number of args to 256. This restriction may
 * be removed if needed in the future, with more elaborate reallocing code.
 */

int					/* todo: K&R definition */
SES_DeclareFunction(SES_Connection sescon, char *name,
		    void (*func)(SES_Args args), int arity, int return_type, ...)
{
    int namelen = strlen(name);
    SES_PS_Header request;
    uchar *ptr;
    int types[256];
    int n_types = 0, i;
    static int funcnum;
    va_list argList;

    va_start(argList, return_type);
    
    while (types[n_types] = va_arg(argList, int)) {
	if (++n_types >= 256)
	    SES_FatalIOErrorHandler(sescon, 8);
    }
    request = SES__NewRequest(sescon, SES_HT_In_Send, 
			      SES_HFI_DeclareFunction, 
			      0,
			      SES_PS_Ptr_padlen
			      + SES_PS_String_padlen(namelen)
			      + SES_PS_Int_padlen
			      + SES_PS_Int_padlen
			      + n_types * SES_PS_Int_padlen,
			      4 + n_types);
    ptr = SES_PS_HeaderEnd(request);
    ptr = SES__EncodePtr(ptr, func);
    ptr = SES__EncodeString(ptr, namelen, name);
    ptr = SES__EncodeInt(ptr, arity);
    ptr = SES__EncodeInt(ptr, return_type);
    for (i = 0; i < n_types; i++)
	ptr = SES__EncodeInt(ptr, types[i]);

    return ++funcnum;			/* returns number of function (unused) */
}

/*****************************************************************************\
* 				 Declare Type                                 *
\*****************************************************************************/
/* Declaring a type is sending the built-in function SES_HFI_DeclareType
 * with three parameters:
 *   - ID of this type (number)
 *   - name of this type (string)
 *   - close callback
 */  

int 
SES_DeclareType(sescon, name, freecallback)
    SES_Connection sescon;
    char *name;
    void (*freecallback)();
{
    int namelen = strlen(name);
    SES_PS_Header request = SES__NewRequest(sescon, SES_HT_In_Send, 
					    SES_HFI_DeclareType, 0,
					    2 * SES_PS_Int_padlen
					    + SES_PS_String_padlen(namelen),
					    3);
    uchar *ptr = SES_PS_HeaderEnd(request);
    ptr = SES__EncodeInt(ptr, SES__id_last);
    ptr = SES__EncodeString(ptr, namelen, name);
    ptr = SES__EncodePtr(ptr, (void *) freecallback);
    /* todo: we could cache locally the callbacks later to optimize */
    return SES__id_last++;
}

/*****************************************************************************\
* 			       Parses Arguments                               *
\*****************************************************************************/

int 
SES_ArgTypeOf(SES_Args args)
{
    return ((SES_PS_Any) (args->ptr))->type;
}

int 
SES_ArgNumberOf(SES_Args args)
{
    return args->request->arity - args->position;
}

/* Pattern:
 *   - check argument type (may be not necessary actually)
 *   - increment position
 *   - set result
 *   - advance char pointer in arg stream (with padded lengths)
 *   - return result
 */

#define SES__ArgMustBe(sestype, id, ctype, field) 		\
    ctype result;					\
    if (((sestype) (args->ptr))->type != id)		\
	SES_FatalIOErrorHandler(args->sescon, 6);	\
    args->position++;					\
    result = (ctype) ((sestype) (args->ptr))->field;
       
int 
SES_ArgInt(SES_Args args)
{
    SES__ArgMustBe(SES_PS_Int, SES_Int, int, number);
    args->ptr += SES_PS_Int_padlen;
    return result;
}

char *
SES_ArgString(SES_Args args)
{
    SES__ArgMustBe(SES_PS_String, SES_String, char *, string);
    args->ptr += SES_PS_String_padlen(((SES_PS_String) (args->ptr))->length);
    return result;
}

/* same as above, but returns also length in second argument */
char *
SES_ArgStringAndLength(SES_Args args, int *lengthp)
{
    SES__ArgMustBe(SES_PS_String, SES_String, char *, string);
    *lengthp = ((SES_PS_String) (args->ptr))->length;
    args->ptr += SES_PS_String_padlen(((SES_PS_String) (args->ptr))->length);
    return result;
}

char *
SES_ArgAtom(SES_Args args)
{
    SES__ArgMustBe(SES_PS_Atom, SES_Atom, char *, string);
    args->ptr += SES_PS_Atom_padlen(((SES_PS_Atom) (args->ptr))->length);
    return result;
}

void *
SES_ArgId(SES_Args args, int idtype)
{
    void *result;
    if ((((SES_PS_Id) (args->ptr))->type != SES_Id)
	|| ((SES_PS_Id) (args->ptr))->idtype != idtype)
	SES_FatalIOErrorHandler(args->sescon, 6);
    args->position++;
    result = ((SES_PS_Id) (args->ptr))->value;
    args->ptr += SES_PS_Id_padlen;
    return result;
}

int 
SES_ArgBoolean(SES_Args args)
{
    SES__ArgMustBe(SES_PS_Boolean, SES_Boolean, int, number);
    args->ptr += SES_PS_Boolean_padlen;
    return result;
}

double
SES_ArgReal(SES_Args args)
{
    SES__ArgMustBe(SES_PS_Real, SES_Real, double, number);
    args->ptr += SES_PS_Real_padlen;
    return result;
}

void *
SES_ArgPtr(SES_Args args)
{
    SES__ArgMustBe(SES_PS_Ptr, SES_Ptr, void *, ptr);
    args->ptr += SES_PS_Ptr_padlen;
    return result;
}

/*****************************************************************************\
* 				Return Results                                *
\*****************************************************************************/
/* Returning a value is putting a request of type SES_HT_In_Result in the 
 * output queue. Actually, this request header is already put (with size 
 * reserved for a Int, if more is needed, adjust), the only thing to do 
 * is to append the value.
 */

void 
SES_ReturnInt(SES_Args args, int n)
{
    if (args->return_result) {		/* else, discard result */
	SES_PS_Int answer = (SES_PS_Int) SES_PS_HeaderEnd(args->return_result);
	/* no length adjusting necessary */
	answer->type = SES_Int;
	answer->number = n;
    }
}

void 
SES_ReturnString(SES_Args args, char *string)
{
    if (args->return_result) {		/* else, discard result */
	int len = strlen(string);
	SES_PS_String answer;
	/* reserve more room */
	args->return_result = 
	    SES__AppendToRequest(args->sescon, args->return_result,
				 SES_PS_String_padlen(len), 0);
	answer = (SES_PS_String) SES_PS_HeaderEnd(args->return_result);
	answer->type = SES_String;
	answer->length = len;
	memcpy(answer->string, string, len+1);
    }
}

void 
SES_ReturnAtom(SES_Args args, char *string)
{
    if (args->return_result) {		/* else, discard result */
	int len = strlen(string);
	SES_PS_Atom answer;
	/* reserve more room */
	args->return_result = 
	    SES__AppendToRequest(args->sescon, args->return_result,
				 SES_PS_Atom_padlen(len), 0);
	answer = (SES_PS_Atom) SES_PS_HeaderEnd(args->return_result);
	answer->type = SES_Atom;
	answer->length = len;
	memcpy(answer->string, string, len+1);
    }
}

void 
SES_ReturnId(SES_Args args, int idtype, void *value)
{
    if (args->return_result) {		/* else, discard result */
	SES_PS_Id answer;
	/* reserve more room */
	args->return_result = 
	    SES__AppendToRequest(args->sescon, args->return_result,
				 SES_PS_Id_padlen - SES_PS_Int_padlen, 0);
	answer = (SES_PS_Id) SES_PS_HeaderEnd(args->return_result);
	answer->type = SES_Id;
	answer->idtype = idtype;
	answer->value = value;
    }
}

void 
SES_ReturnBoolean(SES_Args args, int n)
{
    if (args->return_result) {		/* else, discard result */
	SES_PS_Boolean answer = (SES_PS_Boolean) 
	    SES_PS_HeaderEnd(args->return_result);
	/* no length adjusting necessary */
	answer->type = SES_Boolean;
	answer->number = n;
    }
}

void
SES_ReturnReal(SES_Args args, double n)
{
    if (args->return_result) {		/* else, discard result */
	SES_PS_Real answer = (SES_PS_Real) 
	    SES_PS_HeaderEnd(args->return_result);
	args->return_result = 
	    SES__AppendToRequest(args->sescon, args->return_result,
				 SES_PS_Real_padlen - SES_PS_Int_padlen, 0);
	answer->type = SES_Real;
	answer->number = n;
    }
}

void
SES_ReturnPtr(SES_Args args, void *ptr)
{
    if (args->return_result) {		/* else, discard result */
	SES_PS_Ptr answer = (SES_PS_Ptr) 
	    SES_PS_HeaderEnd(args->return_result);
	if (sizeof(void *) > sizeof(int)) { /* compiler may discard this */
	    args->return_result =	/* adjust length if necessary */
		SES__AppendToRequest(args->sescon, args->return_result,
				     SES_PS_Ptr_padlen - SES_PS_Int_padlen, 0);
	}
	answer->type = SES_Ptr;
	answer->ptr = ptr;
    }
}

/*****************************************************************************\
* 				 Raise Error                                  *
\*****************************************************************************/

void 			
SES_ReturnERROR(SES_Args args, char *error_argument_list)
{
    int len = strlen(error_argument_list);
    SES_PS_String answer;
    if (!args->return_result) {		/* then create one */
	args->return_result = SES__NewRequest(args->sescon, SES_HT_In_Error,
					      0,
					      args->request->reqid,
					      SES_PS_Int_padlen, /* guess */
					      1);
    }
    args->return_result->type = SES_HT_In_Error;
    /* reserve more room */
    args->return_result = SES__AppendToRequest(args->sescon, args->return_result,
					       SES_PS_String_padlen(len), 0);
    answer = (SES_PS_String) SES_PS_HeaderEnd(args->return_result);
    answer->type = SES_String;
    answer->length = len;
    memcpy(answer->string, error_argument_list, len+1);
}

/*****************************************************************************\
* 				  Send Text                                   *
\*****************************************************************************/
/* sends Klone text to be executed by the main body. We may want to guard
 * the evaluation by some safety mecanism later on?
 */

int 
SES_SendText(sescon, callback, text, length, closure, blocking)
    SES_Connection sescon;
    void (*callback)();
    char *text;
    int length;
    void *closure;
    int blocking;
{
    SES_PS_Header request = SES__NewRequest(sescon, SES_HT_In_Send, 
					    SES_HFI_ReadEvalText, 0,
					    SES_PS_String_padlen(length),
					    1);
    uchar *ptr = SES_PS_HeaderEnd(request);
    ptr = SES__EncodeString(ptr, length, text);

    if (callback) {
	/* TODO */
	/* set up closure to trap the SES_HT_Out_Result or SES_HT_Out_Error 
	   answer */
    }
    return 0;
}

/*****************************************************************************\
* 			       Init Connection                                *
\*****************************************************************************/
/* sets up the connection. returns a connection object.
 * In case of errors, aborts via SES_FatalIOErrorHandler (redefinable)
 * 
 * the connection is described by the value of the shell variable KLONE_SES_ARGS
 * which is a set of 8 numbers:
 * [0] the version of the protocol (now 0)
 * [1] the transport medium type: 0 = pipe, shared memory = 1, same_process = 2
 * [2] key of the IPC queue (shm)
 * [3] key of the IPC semaphores (shm)
 * [4] file descriptor main->extension (unique)
 * [5] file descriptor extension->main (shared)
 * [6] the process ID of the parent application
 * [7] ID (byte) of the extension
 */

SES_Connection				/* the connection object */
SES_Init(name)
    char *name;				/* the name of the module */
{
    char *env_parms = getenv("KLONE_SES_ARGS");
    SES_Connection sescon = (SES_Connection) 
	malloc(sizeof(struct _SES_Connection));
    
#ifdef SESDEBUG
    fprintf(stderr, "SESLIB: extension %s started with %s\n", 
	    name ? name : "(null)", 
	    env_parms ? env_parms : "no KLONE_SES_ARGS variable");
    fflush(stderr);
#endif

    /* decode parameters from father */
    if (!env_parms 
	|| (*env_parms == '\0')
	|| (sscanf(env_parms, "%d:%d:%d:%d:%d:%d:%d:%d",
		   &(sescon->version), &(sescon->transport), &(sescon->key_shm),
		   &(sescon->key_sem), &(sescon->fd_in), &(sescon->fd_out), 
		   &(sescon->pid), &(sescon->id))
	    != 8)) {
	free(sescon);
	SES_errno = 1;
	SES_FatalIOErrorHandler(sescon, 1);
    }
    /* sanity check on parameters */

    /* global data inits */
    SES_ConnectionInit(sescon);

    /* put file descriptor main->extension in non-blocking mode */
    {
	int file_mode;
	fcntl(sescon->fd_in, F_GETFL, &file_mode);
	file_mode |= KlNDELAY;	/* set non-blocking flag */
	fcntl(sescon->fd_in, F_SETFL, file_mode);
    }

    return SES__CurrentSescon = sescon;
}

void
SES_Ready(SES_Connection sescon)
{
    SES__NewRequest(sescon, SES_HT_In_Send, SES_HFI_Ready, 0, 0, 0);
}


/*****************************************************************************\
* 				   mainloop                                   *
\*****************************************************************************/
/* simple  mainloop
 * todo: implements as hooks to the standard one 
 */

void
SES_MainLoop(SES_Connection sescon)
{
    fd_set fds;

    SES_Ready(sescon);

    for (;;) {
	FD_ZERO(&fds);
	FD_SET(SES_ConnectionFD(sescon), &fds);
	SES_MainPreSelect(sescon);
	select(SES_ConnectionFD(sescon) + 1, &fds, NULL, NULL, NULL);
	if (FD_ISSET(SES_ConnectionFD(sescon), &fds) > 0) {
	    SES_Handler(sescon);
	} else {
	    exit(0);			/* should not happen */
	}
    }
}

/* if you redefine the main loop, you MUST call this before entering the select
 */

void
SES_MainPreSelect(SES_Connection sescon)
{
    if (sescon->outptr > sescon->outbuf)
	SES_Flush(sescon);
    while (SES__RequestPending(sescon)) {
	SES__DecodeRequest(sescon, (SES_PS_Header) sescon->inptr);
    }
}
    

SES_InputId
SES_AddInput(SES_Connection sescon,
	     int source,
	     int mask,
	     SES_Callback callback,
	     void *client_data)
{
    /* TODO */
}

void
SES_RemoveInput(SES_InputId inputid)
{
    /* TODO */
}

SES_TimerId
SES_AddTimeOut(SES_Connection sescon,
	     unsigned long interval,
	     SES_TimerCallback callback,
	     void *client_data)
{
    /* TODO */
}

void
SES_RemoveTimeOut(SES_TimerId timerid)
{
    /* TODO */
}

/*****************************************************************************\
* 				mainloop logic                                *
\*****************************************************************************/
/* send error, then wait for the Klone side error handler to decide what to
 * do: proceed with next requests, or discard?
 * which is done by sending a
 */

void
SES__WaitForErrorHandling(SES_Args args)
{
    args->sescon->errhandler_reply = 0;
    SES_Flush(args->sescon);
    while (!args->sescon->errhandler_reply) {
	SES_ReadAndHandle(args->sescon);
    }
    if (args->sescon->errhandler_reply == 2) { /* discards queue */
	/* todo: discard pending events */
    }
}
